#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""***********************************************************************************
  Benchmark 1
  Also available at http://www.daetools.com/docs/tutorials-cv.html#tutorial-cv-4
************************************************************************************"""
import sys, numpy
from time import localtime, strftime
import matplotlib.pyplot as plt
from daetools.pyDAE import *
from daetools.solvers.superlu import pySuperLU
from daetools.pyDAE.evaluator_opencl import pyEvaluator_OpenCL

no_t = daeVariableType("no_t", dimless, -1.0e+20, 1.0e+20, 0.0, 1e-6)

class modTutorial(daeModel):
    def __init__(self, Name, Parent = None, Description = ""):
        daeModel.__init__(self, Name, Parent, Description)

        self.x  = daeDomain("x", self, m, "X axis domain")
        self.y  = daeDomain("y", self, m, "Y axis domain")

        self.u0  = 1.0
        self.v0  = 1.0
        self.w0  = 0.1
        self.ni  = 0.7
        self.eps = 0.001

        self.u  = daeVariable("u",  no_t, self, "", [self.x, self.y])
        self.v  = daeVariable("v",  no_t, self, "", [self.x, self.y])
        self.um = daeVariable("um", no_t, self, "", [self.x, self.y])
        self.vm = daeVariable("vm", no_t, self, "", [self.x, self.y])

    def DeclareEquations(self):
        daeModel.DeclareEquations(self)

        # Create some auxiliary functions to make equations more readable 
        u0       = self.u0
        v0       = self.v0
        ni       = self.ni
        w0       = self.w0
        eps      = self.eps
        t       = Time()
        
        u       = lambda x,y: self.u(x,y)
        v       = lambda x,y: self.v(x,y)
        du_dt   = lambda x,y: dt(u(x,y))
        duu_dx  = lambda x,y:  d(u(x,y)*u(x,y), self.x)
        duv_dy  = lambda x,y:  d(u(x,y)*v(x,y), self.y)
        d2u_dx2 = lambda x,y: d2(u(x,y), self.x)
        d2u_dy2 = lambda x,y: d2(u(x,y), self.y)
        
        dv_dt   = lambda x,y: dt(v(x,y))
        dvu_dx  = lambda x,y:  d(v(x,y)*u(x,y), self.x)
        dvv_dy  = lambda x,y:  d(v(x,y)*v(x,y), self.y)
        d2v_dx2 = lambda x,y: d2(v(x,y), self.x)
        d2v_dy2 = lambda x,y: d2(v(x,y), self.y)

        um       = lambda x,y: u0 * (numpy.sin(x()**2 + y()**2 + w0*t) + eps)
        dum_dt   = lambda x,y: dt(um(x,y))
        dumum_dx = lambda x,y:  d(um(x,y)*um(x,y), self.x)
        dumvm_dy = lambda x,y:  d(um(x,y)*vm(x,y), self.y)
        d2um_dx2 = lambda x,y: d2(um(x,y), self.x)
        d2um_dy2 = lambda x,y: d2(um(x,y), self.y)
        
        vm       = lambda x,y: v0 * (numpy.cos(x()**2 + y()**2 + w0*t) + eps)
        dvm_dt   = lambda x,y: dt(vm(x,y))
        dvmum_dx = lambda x,y:  d(vm(x,y)*um(x,y), self.x)
        dvmvm_dy = lambda x,y:  d(vm(x,y)*vm(x,y), self.y)
        d2vm_dx2 = lambda x,y: d2(vm(x,y), self.x)
        d2vm_dy2 = lambda x,y: d2(vm(x,y), self.y)
        
        Su       = lambda x,y: dum_dt(x,y) + (dumum_dx(x,y) + dumvm_dy(x,y)) - ni * (d2um_dx2(x,y) + d2um_dy2(x,y))
        Sv       = lambda x,y: dvm_dt(x,y) + (dvmum_dx(x,y) + dvmvm_dy(x,y)) - ni * (d2vm_dx2(x,y) + d2vm_dy2(x,y))

        # Numerical solution
        eq = self.CreateEquation("u", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eOpenOpen)
        y = eq.DistributeOnDomain(self.y, eOpenOpen)
        eq.Residual = du_dt(x,y) + (duu_dx(x,y) + duv_dy(x,y)) - ni * (d2u_dx2(x,y) + d2u_dy2(x,y)) - Su(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("u(,0)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eOpenOpen)
        y = eq.DistributeOnDomain(self.y, eLowerBound)
        eq.Residual = u(x,y) - um(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("u(,1)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eOpenOpen)
        y = eq.DistributeOnDomain(self.y, eUpperBound)
        eq.Residual = u(x,y) - um(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("u(0,)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eLowerBound)
        y = eq.DistributeOnDomain(self.y, eClosedClosed)
        eq.Residual = u(x,y) - um(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("u(1,)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eUpperBound)
        y = eq.DistributeOnDomain(self.y, eClosedClosed)
        eq.Residual = u(x,y) - um(x,y)
        eq.CheckUnitsConsistency = False


        # v component
        eq = self.CreateEquation("v", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eOpenOpen)
        y = eq.DistributeOnDomain(self.y, eOpenOpen)
        eq.Residual = dv_dt(x,y) + (dvu_dx(x,y) + dvv_dy(x,y)) - ni * (d2v_dx2(x,y) + d2v_dy2(x,y)) - Sv(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("v(,0)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eOpenOpen)
        y = eq.DistributeOnDomain(self.y, eLowerBound)
        eq.Residual = v(x,y) - vm(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("v(,1)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eOpenOpen)
        y = eq.DistributeOnDomain(self.y, eUpperBound)
        eq.Residual = v(x,y) - vm(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("v(0,)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eLowerBound)
        y = eq.DistributeOnDomain(self.y, eClosedClosed)
        eq.Residual = v(x,y) - vm(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("v(1,)", "Numerical solution")
        x = eq.DistributeOnDomain(self.x, eUpperBound)
        y = eq.DistributeOnDomain(self.y, eClosedClosed)
        eq.Residual = v(x,y) - vm(x,y)
        eq.CheckUnitsConsistency = False

        # Manufactured solution
        eq = self.CreateEquation("um", "Manufactured solution")
        x = eq.DistributeOnDomain(self.x, eClosedClosed)
        y = eq.DistributeOnDomain(self.y, eClosedClosed)
        eq.Residual = self.um(x,y) - um(x,y)
        eq.CheckUnitsConsistency = False

        eq = self.CreateEquation("vm", "Manufactured solution")
        x = eq.DistributeOnDomain(self.x, eClosedClosed)
        y = eq.DistributeOnDomain(self.y, eClosedClosed)
        eq.Residual = self.vm(x,y) - vm(x,y)
        eq.CheckUnitsConsistency = False

class simTutorial(daeSimulation):
    def __init__(self, Nx, Ny):
        daeSimulation.__init__(self)
        self.m = modTutorial("benchmark1(%dx%d)" % (Nx,Ny))
        
        self.Nx = Nx
        self.Ny = Ny

    def SetUpParametersAndDomains(self):
        self.m.x.CreateStructuredGrid(self.Nx, -0.1, 0.7)
        self.m.y.CreateStructuredGrid(self.Ny,  0.2, 0.8)
        
    def SetUpVariables(self):
        Nx = self.m.x.NumberOfPoints
        Ny = self.m.y.NumberOfPoints
        
        xp = self.m.x.Points
        yp = self.m.y.Points
        
        u0       = self.m.u0
        v0       = self.m.v0
        eps      = self.m.eps
        
        um0 = lambda x,y: u0 * (numpy.sin(x**2 + y**2) + eps)
        vm0 = lambda x,y: v0 * (numpy.cos(x**2 + y**2) + eps)
        
        for x in range(1, Nx-1):
            for y in range(1, Ny-1):
                self.m.u.SetInitialCondition(x,y, um0(xp[x], yp[y]))
                self.m.v.SetInitialCondition(x,y, vm0(xp[x], yp[y]))
                
# Setup everything manually and run in a console
def run(Nx, Ny, **kwargs):
    # Create Log, Solver, DataReporter and Simulation object
    log          = daePythonStdOutLog()
    daesolver    = daeIDAS()
    simulation   = simTutorial(Nx, Ny)

    # Do no print progress
    log.PrintProgress = False

    # Create SuperLU LA solver if not sent as a keyword argument
    if 'lasolver' in kwargs:
        lasolver = kwargs['lasolver']
        del kwargs['lasolver']
    else:
        lasolver = pySuperLU.daeCreateSuperLUSolver()

    # Enable reporting of time derivatives for all reported variables
    simulation.ReportTimeDerivatives = True

    if 'datareporter' in kwargs:
        datareporter = kwargs['datareporter']
        del kwargs['datareporter']
    else:
        datareporter = daeDelegateDataReporter()
        # 1. TCP/IP
        tcpipDataReporter = daeTCPIPDataReporter()
        datareporter.AddDataReporter(tcpipDataReporter)
        simName = simulation.m.Name + strftime(" [%d.%m.%Y %H:%M:%S]", localtime())
        if not tcpipDataReporter.Connect("", simName):
            sys.exit()

        # 2. Data
        dr = daeNoOpDataReporter()
        datareporter.AddDataReporter(dr)

    return daeActivity.simulate(simulation, reportingInterval = 2.0, 
                                            timeHorizon       = 90.0,
                                            lasolver          = lasolver,
                                            datareporter      = datareporter,
                                            **kwargs)    

if __name__ == "__main__":
    Nx = 80
    Ny = 64
    
    cfg = daeGetConfig()
    
    # Case 1. CS Sequential
    #cfg.SetString('daetools.core.equations.evaluationMode', 'computeStack_OpenMP')
    #cfg.SetInteger('daetools.core.equations.computeStack_OpenMP.numThreads', 1)

    # Case 2. CS OpenMP (using the default number of threads = Ncores)
    #cfg.SetString('daetools.core.equations.evaluationMode', 'computeStack_OpenMP')
    #cfg.SetInteger('daetools.core.equations.computeStack_OpenMP.numThreads', 0)

    # Case 3. ET Sequential
    #cfg.SetString('daetools.core.equations.evaluationMode', 'evaluationTree_OpenMP')
    #cfg.SetInteger('daetools.core.equations.evaluationTree_OpenMP.numThreads', 1)

    # Case 4. ET OpenMP (using the default number of threads = Ncores)
    #cfg.SetString('daetools.core.equations.evaluationMode', 'evaluationTree_OpenMP')
    #cfg.SetInteger('daetools.core.equations.evaluationTree_OpenMP.numThreads', 0)

    # Cases 5 - 12
    # Available OpenCL platforms/devices can be obtained using the following functions: 
    #openclPlatforms = pyEvaluator_OpenCL.AvailableOpenCLPlatforms()
    #openclDevices   = pyEvaluator_OpenCL.AvailableOpenCLDevices()
    #print('Available OpenCL platforms:')
    #for platform in openclPlatforms:
    #    print('  Platform: %s' % platform.Name)
    #    print('    PlatformID: %d' % platform.PlatformID)
    #    print('    Vendor:     %s' % platform.Vendor)
    #    print('    Version:    %s' % platform.Version)
    #    #print('    Profile:    %s' % platform.Profile)
    #    #print('    Extensions: %s' % platform.Extensions)
    #    print('')
    #print('Available OpenCL devices:')
    #for device in openclDevices:
    #    print('  Device: %s' % device.Name)
    #    print('    PlatformID:      %d' % device.PlatformID)
    #    print('    DeviceID:        %d' % device.DeviceID)
    #    print('    DeviceVersion:   %s' % device.DeviceVersion)
    #    print('    DriverVersion:   %s' % device.DriverVersion)
    #    print('    OpenCLVersion:   %s' % device.OpenCLVersion)
    #    print('    MaxComputeUnits: %d' % device.MaxComputeUnits)
    #    print('')

    cse = None
    # OpenCL evaluators can use a single or multiple OpenCL devices.
    #   a) Single OpenCL device:
    #cse = pyEvaluator_OpenCL.CreateComputeStackEvaluator(platformID = 0, deviceID = 0)
    #   b) Multiple OpenCL devices (for heterogeneous computing) use a list of tuples:
    #      (platformID, deviceID, fractionOfTotalNumberOfEquationsToBeEvaluated)
    #cse = pyEvaluator_OpenCL.CreateComputeStackEvaluator( [(0, 0, 0.6), (1, 1, 0.4)] )
    
    run(Nx, Ny, computeStackEvaluator = cse)
