/***********************************************************************************
                 DAE Tools Project: www.daetools.com
                 Copyright (C) Dragan Nikolic
************************************************************************************
DAE Tools is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License version 3 as published by the Free Software
Foundation. DAE Tools is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with the
DAE Tools software; if not, see <http://www.gnu.org/licenses/>.
***********************************************************************************/
#include "model.h"
#include <exception>
#include <fstream>
#include <iomanip>
#include "compute_stack_openmp.h"
#include "benchmark_1_cxx.h"

namespace daetools_model
{
benchmark1 benchmark1_model;

/* Internal storage containers for evaluation of equations. */
std::vector<real_t>  g_values;
std::vector<real_t>  g_timeDerivatives;
std::vector<real_t>  g_jacobian_cxx;
std::vector<real_t>  g_jacobian_cs;
std::vector<real_t>  g_residuals_cxx;
std::vector<real_t>  g_residuals_cs;

runtimeInformationData_t g_rtData;

/* ComputeStack-related data. */
std::vector<adComputeStackItem_t>   g_arrAllComputeStacks;
std::vector<adJacobianMatrixItem_t> g_arrJacobianMatrixItems;
std::vector<uint32_t>               g_arrActiveEquationSetIndexes;
uint32_t*                           g_activeEquationSetIndexes  = NULL;
adComputeStackItem_t*               g_computeStacks             = NULL;
adJacobianMatrixItem_t*             g_jacobianMatrixItems       = NULL;
uint32_t                            g_numberOfJacobianItems     = -1;
uint32_t                            g_numberOfComputeStackItems = -1;
adComputeStackEvaluator_t*          g_pEvaluator                = NULL;

static void loadRuntimeData(const std::string& inputDirectory);
static void loadModelEquations(const std::string& inputDirectory);
static void loadJacobianData(const std::string& inputDirectory);

void LoadModel(const std::string& inputDirectory)
{
    loadRuntimeData(inputDirectory);

    loadModelEquations(inputDirectory);

    loadJacobianData(inputDirectory);

    g_values          = g_rtData.init_values;
    g_timeDerivatives = g_rtData.init_derivatives;

    g_jacobian_cxx.resize(g_numberOfJacobianItems, 0.0);
    g_jacobian_cs.resize(g_numberOfJacobianItems, 0.0);
    g_residuals_cxx.resize(g_rtData.Nequations_local, 0.0);
    g_residuals_cs.resize(g_rtData.Nequations_local, 0.0);

    if(benchmark1_model.Nequations != g_rtData.Nequations_local)
        throw std::runtime_error("Invalid number of equations");

    g_pEvaluator = new daeComputeStackEvaluator_OpenMP();
    g_pEvaluator->Initialize(false,
                             g_rtData.Nequations_local,
                             g_rtData.Nequations_local,
                             g_rtData.Ndofs,
                             g_numberOfComputeStackItems,
                             g_numberOfJacobianItems,
                             g_numberOfJacobianItems,
                             g_computeStacks,
                             g_activeEquationSetIndexes,
                             g_jacobianMatrixItems);
}

void FreeModel()
{
    if(g_pEvaluator)
    {
        g_pEvaluator->FreeResources();
        delete g_pEvaluator;
        g_pEvaluator = NULL;
    }
}

void CompareResiduals()
{
    printf("residuals:\n");
    for(int i = 0; i < g_residuals_cxx.size(); i++)
        printf("%5d: %25.15f\n", i, g_residuals_cxx[i] - g_residuals_cs[i]);
}

void CompareJacobian()
{
    printf("Jacobian:\n");
    for(int i = 0; i < g_jacobian_cxx.size(); i++)
        printf("%5d: %25.15f\n", i, g_jacobian_cxx[i] - g_jacobian_cs[i]);
}

void EvaluateResiduals_cxx()
{
    real_t  current_time     = 10.0;
    real_t* pdofs            = NULL;
    real_t* pvalues          = &g_values[0];
    real_t* ptimeDerivatives = &g_timeDerivatives[0];

    benchmark1_model.EvaluateResiduals(current_time, pvalues, ptimeDerivatives, &g_residuals_cxx[0]);
}

void EvaluateResiduals_cs()
{
    real_t  current_time     = 10.0;
    real_t* pdofs            = NULL;
    real_t* pvalues          = &g_values[0];
    real_t* ptimeDerivatives = &g_timeDerivatives[0];

    daeComputeStackEvaluationContext_t EC;
    EC.equationCalculationMode    = eCalculate;
    EC.sensitivityParameterIndex  = -1;
    EC.jacobianIndex              = -1;
    EC.numberOfVariables          = g_rtData.Nequations_local;
    EC.numberOfEquations          = g_rtData.Nequations_local; // ???
    EC.numberOfDOFs               = 0;
    EC.numberOfComputeStackItems  = g_numberOfComputeStackItems;
    EC.numberOfJacobianItems      = 0;
    EC.valuesStackSize            = 5;
    EC.lvaluesStackSize           = 20;
    EC.rvaluesStackSize           = 5;
    EC.currentTime                = current_time;
    EC.inverseTimeStep            = 0; // Should not be needed here. Double check...
    EC.startEquationIndex         = 0; // !!!
    EC.startJacobianIndex         = 0; // !!!

    g_pEvaluator->EvaluateResiduals(EC, pdofs, pvalues, ptimeDerivatives, &g_residuals_cs[0]);
}

void EvaluateJacobian_cxx()
{
    real_t  current_time      = 10.0;
    real_t  inverse_time_step = 1e5;
    real_t* pdofs             = NULL;
    real_t* pvalues           = &g_values[0];
    real_t* ptimeDerivatives  = &g_timeDerivatives[0];

    benchmark1_model.EvaluateJacobian(current_time, inverse_time_step, pvalues, ptimeDerivatives, &g_jacobian_cxx[0]);
}

void EvaluateJacobian_cs()
{
    real_t  current_time      = 10.0;
    real_t  inverse_time_step = 1e5;
    real_t* pdofs             = NULL;
    real_t* pvalues           = &g_values[0];
    real_t* ptimeDerivatives  = &g_timeDerivatives[0];

    daeComputeStackEvaluationContext_t EC;
    EC.equationCalculationMode    = eCalculateJacobian;
    EC.sensitivityParameterIndex  = -1;
    EC.jacobianIndex              = -1;
    EC.numberOfVariables          = g_rtData.Nequations_local;
    EC.numberOfEquations          = g_rtData.Nequations_local; // ???
    EC.numberOfDOFs               = 0;
    EC.numberOfComputeStackItems  = g_numberOfComputeStackItems;
    EC.numberOfJacobianItems      = g_numberOfJacobianItems;
    EC.valuesStackSize            = 5;
    EC.lvaluesStackSize           = 20;
    EC.rvaluesStackSize           = 5;
    EC.currentTime                = current_time;
    EC.inverseTimeStep            = inverse_time_step;
    EC.startEquationIndex         = 0; // !!!
    EC.startJacobianIndex         = 0; // !!!

    g_pEvaluator->EvaluateJacobian(EC, pdofs, pvalues, ptimeDerivatives, &g_jacobian_cs[0]);
}


void loadRuntimeData(const std::string& inputDirectory)
{
    std::ifstream f;

    std::string filePath = inputDirectory + "/" + "runtime_data.bin";
    f.open(filePath, std::ios_base::binary);
    if(!f.is_open())
        throw std::runtime_error("Cannot open " + filePath + " file");

    f.read((char*)&g_rtData.Ntotal_vars,        sizeof(uint32_t));
    f.read((char*)&g_rtData.Nequations,         sizeof(uint32_t));
    f.read((char*)&g_rtData.Nequations_local,   sizeof(uint32_t));
    f.read((char*)&g_rtData.Ndofs,              sizeof(uint32_t));

    f.read((char*)&g_rtData.startTime,          sizeof(real_t));
    f.read((char*)&g_rtData.timeHorizon,        sizeof(real_t));
    f.read((char*)&g_rtData.reportingInterval,  sizeof(real_t));
    f.read((char*)&g_rtData.relativeTolerance,  sizeof(real_t));

    f.read((char*)&g_rtData.quasiSteadyState,   sizeof(bool));

    int32_t Nitems;

    // dofs
    f.read((char*)&Nitems,  sizeof(int32_t));
    if(Nitems > 0)
    {
        g_rtData.dofs.resize(Nitems);
        f.read((char*)(&g_rtData.dofs[0]), sizeof(real_t) * Nitems);
    }

    // init_values
    f.read((char*)&Nitems,  sizeof(int32_t));
    if(Nitems > 0)
    {
        g_rtData.init_values.resize(Nitems);
        f.read((char*)(&g_rtData.init_values[0]), sizeof(real_t) * Nitems);
    }

    // init_derivatives
    f.read((char*)&Nitems,  sizeof(int32_t));
    if(Nitems > 0)
    {
        g_rtData.init_derivatives.resize(Nitems);
        f.read((char*)(&g_rtData.init_derivatives[0]), sizeof(real_t) * Nitems);
    }

    // absolute_tolerances
    f.read((char*)&Nitems,  sizeof(int32_t));
    if(Nitems > 0)
    {
        g_rtData.absolute_tolerances.resize(Nitems);
        f.read((char*)(&g_rtData.absolute_tolerances[0]), sizeof(real_t) * Nitems);
    }

    // ids
    f.read((char*)&Nitems,  sizeof(int32_t));
    if(Nitems > 0)
    {
        g_rtData.ids.resize(Nitems);
        f.read((char*)(&g_rtData.ids[0]), sizeof(int32_t) * Nitems);
    }

    // variable_names (skipped)
    f.read((char*)&Nitems,  sizeof(int32_t));
    if(Nitems == 0)
    {
        g_rtData.variable_names.resize(g_rtData.Nequations_local);
        for(int i = 0; i < g_rtData.Nequations_local; i++)
        {
            g_rtData.variable_names[i] = "";
        }
    }
    else
    {
        g_rtData.variable_names.resize(Nitems);

        int32_t length;
        char name[4096];
        for(int i = 0; i < Nitems; i++)
        {
            // Read string length
            f.read((char*)&length,  sizeof(int32_t));

            // Read string
            f.read((char*)(&name[0]), sizeof(char) * length);
            name[length] = '\0';

            g_rtData.variable_names[i] = std::string(name);
        }
    }

    f.close();
}

void loadModelEquations(const std::string& inputDirectory)
{
    std::ifstream f;

    std::string filePath = inputDirectory + "/" + "model_equations.bin";
    f.open(filePath, std::ios_base::binary);
    if(!f.is_open())
        throw std::runtime_error("Cannot open " + filePath + " file");

    uint32_t Ncs  = 0;
    uint32_t Nasi = 0;

    f.read((char*)&Nasi, sizeof(uint32_t));
    f.read((char*)&Ncs,  sizeof(uint32_t));

    g_arrActiveEquationSetIndexes.resize(Nasi);
    g_arrAllComputeStacks.resize(Ncs);

    f.read((char*)(&g_arrActiveEquationSetIndexes[0]), sizeof(uint32_t)             * Nasi);
    f.read((char*)(&g_arrAllComputeStacks[0]),         sizeof(adComputeStackItem_t) * Ncs);
    f.close();

    g_activeEquationSetIndexes = &g_arrActiveEquationSetIndexes[0];
    g_computeStacks            = &g_arrAllComputeStacks[0];

    g_numberOfComputeStackItems = g_arrAllComputeStacks.size();
}

void loadJacobianData(const std::string& inputDirectory)
{
    std::ifstream f;

    std::string filePath = inputDirectory + "/" + "jacobian_data.bin";
    f.open(filePath, std::ios_base::binary);
    if(!f.is_open())
        throw std::runtime_error("Cannot open " + filePath + " file");

    uint32_t Nji = 0;

    f.read((char*)&Nji, sizeof(uint32_t));

    g_arrJacobianMatrixItems.resize(Nji);

    f.read((char*)(&g_arrJacobianMatrixItems[0]), sizeof(adJacobianMatrixItem_t) * Nji);
    f.close();

    g_jacobianMatrixItems   = &g_arrJacobianMatrixItems[0];

    g_numberOfJacobianItems = g_arrJacobianMatrixItems.size();


    // Populate benchmark1_model jacobianMatrixIndexes map with the same jacobian items.
    benchmark1_model.jacobianMatrixIndexes.resize(g_rtData.Nequations_local);
    for(int ji = 0; ji < g_numberOfJacobianItems; ji++)
    {
        adJacobianMatrixItem_t& jmi = g_arrJacobianMatrixItems[ji];
        benchmark1_model.jacobianMatrixIndexes[jmi.equationIndex].push_back( jmi.blockIndex );
    }
}

}
